/* Define some constants for access on chip resources */
unsigned long int* const PIO_PERC = (unsigned long int* const) 0xfffff800;
unsigned long int* const PIO_OERC = (unsigned long int* const) 0xfffff810;
unsigned long int* const PIO_ODRC = (unsigned long int* const) 0xfffff814;
unsigned long int* const PIO_SODRC = (unsigned long int* const) 0xfffff830;
unsigned long int* const PIO_CODRC = (unsigned long int* const) 0xfffff834;
unsigned long int* const PIO_PDSRC = (unsigned long int* const) 0xfffff83c;

/* Brownout detector related stuff */
unsigned long int* const SUPC_BOMR = (unsigned long int* const) 0xfffffd14;
unsigned long int* const SUPC_SR = (unsigned long int* const) 0xfffffd24;

/* Font table for digits 0-9 and A-F */
/* The segment mapping must match the big LCD lookup table below */
/* Yes, it's an ugly hack, but it does work. */
/* Feel free to do a better one! */
const unsigned char seven_segment_table[]=
{
   0x7b,    /* 0 */
   0x48,    /* 1 */
   0x3d,    /* 2 */
   0x6d,    /* 3 */
   0x4e,    /* 4 */
   0x67,    /* 5 */
   0x77,    /* 6 */
   0x49,    /* 7 */
   0x7f,    /* 8 */
   0x6f,    /* 9 */
   0x5f,    /* A */
   0x76,    /* B */
   0x34,    /* C */
   0x7c,    /* D */
   0x37,    /* E */
   0x17,    /* F */
};

struct lcd_digit_struct
{
   unsigned char address;
   unsigned char bitmask;
};

/* Table of LCD segments 
   Eight entries for each of the 12 main digits (left to right) 
   This table includes segments a-h, but not i, the comma 
   Again, this is an ugly hack and needs to be cleaned up 
   It's an improvement over the last one, but I think I can
   get it down to half this size.
   Appended the exponent digits to the end of the table.
   The exponents don't have decimal points, so I picked an unmapped
   bit so that it wouldn't break the logic.  0x50, 0x20 is valid
   memory, but it doesn't turn on a segment. */
const struct lcd_digit_struct lcd_digit_table[]=
{
   {0x10, 0x7},
   {0x10, 0x6},
   {0x18, 0x7},
   {0x10, 0x8},
   {0x18, 0x6},
   {0x20, 0x6},
   {0x20, 0x7},
   {0x28, 0x8},

   {0x10, 0x9},
   {0x18, 0x8},
   {0x18, 0x9},
   {0x18, 0xa},
   {0x20, 0x8},
   {0x20, 0x9},
   {0x20, 0xa},
   {0x28, 0xa},

   {0x10, 0xb},
   {0x10, 0xa},
   {0x18, 0xb},
   {0x10, 0xc},
   {0x20, 0xb},
   {0x28, 0xb},
   {0x18, 0xc},
   {0x20, 0xc},

   {0x10, 0xe},
   {0x10, 0xd},
   {0x18, 0xd},
   {0x18, 0xe},
   {0x20, 0xd},
   {0x28, 0xd},
   {0x20, 0xe},
   {0x20, 0xf},

   {0x10, 0x10},
   {0x10, 0xf},
   {0x18, 0x10},
   {0x10, 0x11},
   {0x18, 0xf},
   {0x28, 0xf},
   {0x20, 0x10},
   {0x28, 0x11},

   {0x10, 0x12},
   {0x18, 0x11},
   {0x18, 0x12},
   {0x18, 0x13},
   {0x20, 0x11},
   {0x20, 0x12},
   {0x20, 0x13},
   {0x28, 0x13},

   {0x10, 0x14},
   {0x10, 0x13},
   {0x18, 0x14},
   {0x10, 0x15},
   {0x20, 0x14},
   {0x28, 0x14},
   {0x18, 0x15},
   {0x20, 0x15},

   {0x10, 0x17},
   {0x10, 0x16},
   {0x18, 0x16},
   {0x18, 0x17},
   {0x20, 0x16},
   {0x28, 0x16},
   {0x20, 0x17},
   {0x20, 0x18},

   {0x10, 0x19},
   {0x10, 0x18},
   {0x18, 0x19},
   {0x10, 0x1a},
   {0x18, 0x18},
   {0x28, 0x18},
   {0x20, 0x19},
   {0x28, 0x1a},

   {0x10, 0x1b},
   {0x18, 0x1a},
   {0x18, 0x1b},
   {0x18, 0x1c},
   {0x20, 0x1a},
   {0x20, 0x1b},
   {0x20, 0x1c},
   {0x28, 0x1c},

   {0x10, 0x1d},
   {0x10, 0x1c},
   {0x18, 0x1d},
   {0x10, 0x1e},
   {0x20, 0x1d},
   {0x28, 0x1d},
   {0x18, 0x1e},
   {0x20, 0x1e},

   {0x14, 0x0},
   {0x10, 0x1f},
   {0x18, 0x1f},
   {0x1c, 0x0},
   {0x20, 0x1f},
   {0x28, 0x1f},
   {0x24, 0x0},
   {0x2c, 0x1},

   {0x14, 0x2},
   {0x14, 0x1},
   {0x24, 0x2},
   {0x1c, 0x2},
   {0x24, 0x1},
   {0x2c, 0x2},
   {0x2c, 0x3},
   {0x50, 0x20},

   {0x14, 0x4},
   {0x14, 0x3},
   {0x1c, 0x3},
   {0x1c, 0x4},
   {0x24, 0x3},
   {0x2c, 0x4},
   {0x24, 0x4},
   {0x50, 0x20},

   {0x1c, 0x6},
   {0x14, 0x5},
   {0x1c, 0x5},
   {0x24, 0x6},
   {0x24, 0x5},
   {0x2c, 0x5},
   {0x2c, 0x6},
   {0x50, 0x20},
};

/* Place a hex digit (value) at position (digit) on the main 12 digit display */
/* Values > 15 will show as a blank */
void write_char(int digit, int value)
{
   int pattern;
   int bit;
   unsigned long int address;

   /* Font lookup--convert to segments */
   pattern=seven_segment_table[value];

   /* Blank if out of range */
   if (value>15)
      pattern=0;

   /* Look up each segment in the main LCD table and turn 'em on or off */
   for (bit=0; bit<8; bit++)
   {
      /* Grab address of the segment */
      address=0xfffb4200+lcd_digit_table[(digit*8)+bit].address;

      /* Does font call for segment to be on? */
      if (pattern & (1 << bit))
      {
         /* Yep, so look up the bit and set it */
         *(unsigned long int*)address|=(1 << lcd_digit_table[(digit*8)+bit].bitmask);
      }
      else
      {
         /* Segment should be off; lookup and turn if off */
         *(unsigned long int*)address&=~(1 << lcd_digit_table[(digit*8)+bit].bitmask);
      }
   }
}

/* Display a 32 bit value as an 8 digit hex number,
   with leading zeros supressed. 
   The first visible digit will be in position 0, at
   the far left side of the LCD. */
void show_value(unsigned int value)
{
   int digit_pos=0;     /* LCD digit that we're writing */
   int nibble=7;        /* Do the input value one nibble at a time */
   int blank=0;         /* Flag to supress leading zeros */
   int i;               /* Loop counter */

   /* Assuming 32 bits, so do the first seven nibbles */
   for (i=0; i<7; i++)
   {
      /* Is it non-zero? */
      if (((value >> (4*nibble)) & 0x0f)!=0)
      {
         /* Yep, so no more supression */
         blank=1;
      }
      if (blank)
      {
         /* Place the nibble on the screen */
         write_char(digit_pos, ((value >> (4*nibble)) & 0x0f));
         ++digit_pos;
      }
      --nibble;
   }

   /* Last char is always displayed.  After all,
      if you zero supressed 32 bits of zeros, you'd 
      end up with nothing at all on the LCD. */
   write_char(digit_pos++, (value & 0x0f));

   /* Anything not used on the right side needs to be blanked as well. */
   while(digit_pos<8)
   {
      write_char(digit_pos++, 16);
   }
}

/* Turn off all the LCD segments */
void clear_screen()
{
   unsigned long int* i;
   for (i=(unsigned long int*)0xfffb4200; i<(unsigned long int*)0xfffb4250; i++)
   {
      *i=0;
   }
}

/* Kill some time.  Not calibrated, yet. */
void delay(unsigned int x)
{
   volatile int i;
   volatile int j;

   for (i=0; i<x; i++)
   {
      for (j=0; j<1000; j++)
      {
      }
   }
}

/* Hacked up table to convert from brownout threshold to battery voltage.
   Since I don't have a decimal display routine, I just display the
   value as two hex digits, but they're decimal.  In other words, "29"
   would be 2.9V, and "30" would be 3.0V.  */   
const int battery_voltage_table[]=
{
   0x19,
   0x20,
   0x21,
   0x22,
   0x23,
   0x24,
   0x25,
   0x26,
   0x27,
   0x28,
   0x29,
   0x30,
   0x31,
   0x32,
   0x33,
   0x34,
};


/* Continuously probe the brownout threshold and ID when 
   it detects a brownout.  That might be our battery voltage. */
int main(void)
{
   int i;
   int battery_voltage=0;

   clear_screen();

   for(;;)
   {
      /* Step the brownout detector down in voltage until the battery no
         longer sets it off.  When the threshold is less than the battery,
         the brownout will go away and we should have a rough measurement 
         of battery voltage. 

         Example:

            Threshold      Battery     Brownout
            3.2            2.95        yes
            3.1            2.95        yes
            3.0            2.95        yes         <- low water mark
            2.9            2.95        no
            2.8            2.95        no
            2.7            2.95        no

         We can say that the battery voltage is about 3V in this case.
         Right? */
      for (i=0xf; i>=0; i--)
      {
         /* brownout detector runs continously, threshold is i */
         *SUPC_BOMR=0x100 | i;

         /* Do we need to give it time to settle? */
         delay(40);

         /* Check to see if bit 6 has gone high, indicating a brownout */
         if ((*SUPC_SR) & (1 << 6))
         {
            /* yep, brownout.  Mark this as the last low water mark */
            battery_voltage=i;
         }

         /* Show the current probe and last low water mark */

         /* Current brownout threshold setting */
         write_char(9, i);

         /* Last low water mark, moving down as the probe runs */
         write_char(11, battery_voltage);

         /* Display the supply controller status register */
         show_value(*SUPC_SR);
      }

      /* Probing complete, update the exponent area with the battery voltage */
      write_char(12, (battery_voltage_table[battery_voltage] >> 4) & 0x0f);
      write_char(13, (battery_voltage_table[battery_voltage] >> 0) & 0x0f);
   }

   return 0;
}
